import macromol_dataframe as mmdf
import macromol_voxelize as mmvox
import numpy as np
import parametrize_from_file as pff
import pytest

from visible_residues import sample_visible_residues, Sphere
from polars.testing import assert_frame_equal
from pathlib import Path

CIF_DIR = Path(__file__).parent / 'resources'

def grid(params):
    return mmvox.Grid(
            length_voxels=1,
            resolution_A=float(params['length_A']),
            center_A=np.array(eval(params['center_A'])),
    )

def bounding_sphere(params):
    return Sphere(
            center_A=np.array(eval(params['center_A'])),
            radius_A=float(params['radius_A']),
    )


@pytest.mark.parametrize('n', [0, 1, 2])
@pff.parametrize(
        schema=pff.cast(
            grid=grid,
            sphere=bounding_sphere,
            expected=eval,
        ),
)
def test_sample_visible_residues(n, grid, sphere, expected):
    # I deliberately constructed this amino acid structure so that the global 
    # and residue-local coordinate frames are the same, to make it easy to 
    # reason about the expected results.

    atoms = mmdf.read_asymmetric_unit(CIF_DIR / 'axis_aligned_residue.cif')
    atoms = mmdf.assign_residue_ids(atoms)

    visible = sample_visible_residues(
            rng=np.random.default_rng(0),
            atoms=atoms,
            grid=grid,
            n=n,
            bounding_sphere=sphere,
    )
    assert len(visible) == min(expected, n)
    assert (visible['radius_A'] == sphere.radius_A).all()

def test_sample_visible_residues_4rek_G49():
    # This is a residue with two completely separate backbone conformations.  
    # The A conformation has 69% occupancy, so it should always be used, 
    # regardless of the actual order of the atoms in the input file.  We test 
    # by comparing to a file with only the A conformation.

    def sample_visible(alt_ids):
        atoms = mmdf.read_asymmetric_unit(CIF_DIR / f'4rek_G49_{alt_ids}.cif')
        atoms = mmdf.assign_residue_ids(atoms, maintain_order=True)

        # We're not trying to exclude any atoms in this test, so the grid just 
        # needs to fit everything.
        grid = mmvox.Grid(
                length_voxels=15,
                resolution_A=1.0,
                center_A=np.array([0, -5, 30]),
        )

        return sample_visible_residues(
                rng=np.random.default_rng(0),
                atoms=atoms,
                grid=grid,
                n=1,
        )

    visible_A = sample_visible('A')
    visible_AB = sample_visible('AB')
    visible_BA = sample_visible('BA')

    assert len(visible_A) == 1
    assert len(visible_AB) == 1
    assert len(visible_BA) == 1

    assert_frame_equal(visible_A, visible_AB)
    assert_frame_equal(visible_A, visible_BA)

def test_sample_visible_residues_4rek_M154():
    # This is a residue with three Cα conformations, but only one N/C 
    # conformation.  The A conformation has 62% occupancy, so it should always 
    # be used, regardless of the actual order of the atoms in the input file.  
    # We test by comparing to a file with only the A conformation.

    def sample_visible(alt_ids):
        atoms = mmdf.read_asymmetric_unit(CIF_DIR / f'4rek_M154_{alt_ids}.cif')
        atoms = mmdf.assign_residue_ids(atoms, maintain_order=True)

        # We're not trying to exclude any atoms in this test, so the grid just 
        # needs to fit everything.
        grid = mmvox.Grid(
                length_voxels=15,
                resolution_A=1.0,
                center_A=np.array([25, -15, 20]),
        )

        return sample_visible_residues(
                rng=np.random.default_rng(0),
                atoms=atoms,
                grid=grid,
                n=1,
        )

    visible_A = sample_visible('A')
    visible_ABC = sample_visible('ABC')
    visible_CBA = sample_visible('CBA')

    assert len(visible_A) == 1
    assert len(visible_ABC) == 1
    assert len(visible_CBA) == 1

    assert_frame_equal(visible_A.drop('alt_ids'), visible_ABC.drop('alt_ids'))
    assert_frame_equal(visible_A.drop('alt_ids'), visible_CBA.drop('alt_ids'))
    assert_frame_equal(visible_ABC, visible_CBA)

@pytest.mark.parametrize('n', [0, 1, 2, 3, 4, 5])
def test_sample_visible_residues_1lz1(n):
    # The purpose of this test is to make sure that `find_visible_residues` 
    # produces qualitatively reasonable results on a real protein structure.  

    atoms = mmdf.read_asymmetric_unit(CIF_DIR / '1lz1.cif')
    atoms = mmdf.assign_residue_ids(atoms, maintain_order=True)

    grid = mmvox.Grid(length_voxels=15, resolution_A=1.0)

    visible = sample_visible_residues(
            rng=np.random.default_rng(0),
            atoms=atoms,
            grid=grid,
            n=n,
    )
    visible = visible.join(
            atoms.group_by('residue_id', 'seq_id').agg(),
            on='residue_id',
    )

    # These expected sequence ids are based on a manual inspection, see 
    # `resources/1lz1_notes.xlsx` for details.
    assert len(visible) == min(n, 3)
    assert set(visible['seq_id']) <= {35, 57, 96}

@pytest.mark.parametrize('n', [0, 1])
def test_sample_visible_residues_1bna(n):
    # This structure doesn't contain any amino acid residues, but that 
    # shouldn't cause any problems.

    atoms = mmdf.read_asymmetric_unit(CIF_DIR / '1bna.cif')
    atoms = mmdf.assign_residue_ids(atoms, maintain_order=True)

    grid = mmvox.Grid(length_voxels=15, resolution_A=1.0)

    visible = sample_visible_residues(
            rng=np.random.default_rng(0),
            atoms=atoms,
            grid=grid,
            n=n,
    )
    assert len(visible) == 0




import numpy as np
import polars as pl
import macromol_voxelize as mmvox

from ._inner_loop import _find_visible_residues
from macromol_dataframe import Atoms, explode_residue_conformations
from dataclasses import dataclass
from functools import cache

from typing import Optional
from numpy.typing import NDArray

@dataclass
class Sphere:
    center_A: NDArray[float]
    radius_A: float

def sample_visible_residues(
        rng: np.random.Generator,
        atoms: Atoms,
        grid: mmvox.Grid,
        n: int,
        *,
        bounding_sphere: Optional[Sphere] = None,
):
    """
    Find residues for which the bounding sphere is contained entirely within 
    the image.

    Arguments:
        rng:
            A pseudo-random number generator used to ensure that every visible 
            residue has an equal chance of being picked.

        atoms:
            A dataframe of atom coordinates.  The dataframe must contain the 
            following columns:

            - `residue_id`; see `assign_residue_ids()`
            - `atom_id`
            - `x`, `y`, `z`

        img_params:
            The dimensions of the image in question.

        n:
            The maximum number of visible residues to sample.

        bounding_sphere:
            A simplified representation of the region where most sidechain 
            atoms are expected to be found.  See experiment #127 for details.

    Returns:
        A dataframe listing all of residues for which the bounding sphere fits 
        entirely within the image.  Each row also specifies the location of the 
        bounding sphere and the alternate location identifiers of the backbone 
        atoms used to compute the coordinate frame for each residue.

    Note that this function is most efficient for relatively small $n$.
    """

    if bounding_sphere is None:
        bounding_sphere = get_sidechain_bounding_sphere()

    half_boundary_length_A = grid.length_A / 2 - bounding_sphere.radius_A
    min_corner = grid.center_A - half_boundary_length_A
    max_corner = grid.center_A + half_boundary_length_A

    backbone = _select_backbone(atoms)
    backbone = _shuffle_residues(rng, backbone)
    backbone = explode_residue_conformations(backbone, 'alt_id_ex')
    backbone = (
            backbone
            .sort('shuffle_id', 'alt_id_ex', 'backbone_id')
            .rechunk()
    )

    probe_indices = np.zeros(n, dtype=int)
    probe_coords_A = np.zeros((n, 3))

    n_actual = _find_visible_residues(
            backbone['shuffle_id'].to_numpy(allow_copy=False),
            backbone['backbone_id'].to_numpy(allow_copy=False),
            backbone['x'].to_numpy(allow_copy=False),
            backbone['y'].to_numpy(allow_copy=False),
            backbone['z'].to_numpy(allow_copy=False),
            backbone['occupancy'].to_numpy(allow_copy=False),

            bounding_sphere.center_A,
            min_corner,
            max_corner,

            probe_indices,
            probe_coords_A,
    )

    probe_indices = probe_indices[:n_actual]
    probe_coords_A = probe_coords_A[:n_actual]

    probes = pl.DataFrame({
        'residue_id': backbone[probe_indices, 'residue_id'],
        'alt_id_n': backbone[probe_indices, 'alt_id'],
        'alt_id_ca': backbone[probe_indices + 1, 'alt_id'],
        'alt_id_c': backbone[probe_indices + 2, 'alt_id'],
        'x': probe_coords_A[:,0],
        'y': probe_coords_A[:,1],
        'z': probe_coords_A[:,2],
    })

    return probes.select(
            residue_id='residue_id',
            alt_ids=pl.struct(N='alt_id_n', CA='alt_id_ca', C='alt_id_c'),
            x='x',
            y='y',
            z='z',
            radius_A=bounding_sphere.radius_A,
    )

@cache
def get_sidechain_bounding_sphere() -> Sphere:
    # The center coordinate is in a frame aligned with the N, Cα, and C atoms, 
    # and is optimized to include ≈95% of all observed sidechain atoms.  See 
    # experiment #127 for details.
    return Sphere(
            center_A=np.array([
                -1.2729026455200199,
                -1.9322552322708937,
                 2.3466656042276950,
            ]),
            radius_A=4,
    )


def _select_backbone(atoms):
    return (
            atoms
            .filter(
                pl.struct('atom_id', 'element').is_in([
                    dict(atom_id='N', element='N'),
                    dict(atom_id='CA', element='C'),
                    dict(atom_id='C', element='C'),
                ]),
            )
            .with_columns(
                backbone_id=(
                    pl.col('atom_id')
                    .replace_strict({
                        'N': 0,
                        'CA': 1,
                        'C': 2,
                    })
                )
            )
    )

def _shuffle_residues(rng, atoms):
    residue_ids = (
            atoms
            .select('residue_id')
            .unique(maintain_order=True)
    )
    shuffle_ids = (
            residue_ids
            .with_columns(
                shuffle_id=rng.permutation(len(residue_ids)),
            )
    )
    return atoms.join(shuffle_ids, on='residue_id')





import polars as pl

from macromol_dataframe.pymol import set_ascii_dataframe_format
from pymol import cmd, cgo

set_ascii_dataframe_format()

def show_bounding_spheres_cgo(path, radius=4):
    df = pl.read_parquet(path)
    print(df)

    cmd.set('transparency_mode', 1)
    cmd.delete('bounding_spheres')

    #spheres = ['ALPHA', 0.5]
    spheres = []

    for row in df.iter_rows(named=True):
        spheres += [cgo.SPHERE, row['x'], row['y'], row['z'], float(radius)]

    cmd.load_cgo(spheres, 'bounding_spheres')

cmd.extend('show_bounding_spheres_cgo', show_bounding_spheres_cgo)
cmd.auto_arg[0]['show_bounding_spheres_cgo'] = cmd.auto_arg[1]['load']

def show_bounding_spheres_pseudoatom(path):
    df = pl.read_parquet(path)
    print(df)

    cmd.delete('bound')

    for row in df.iter_rows(named=True):
        cmd.pseudoatom(
                'bound',
                resi=row['seq_id'],
                chain=row['chain_id'],
                segi=row['chain_id'],
                pos=[row['x'], row['y'], row['z']],
        )

cmd.extend('show_bounding_spheres_pseudoatom', show_bounding_spheres_pseudoatom)
cmd.auto_arg[0]['show_bounding_spheres_pseudoatom'] = cmd.auto_arg[1]['load']



from setuptools import setup
from pybind11.setup_helpers import Pybind11Extension

setup(
        ext_modules=[
            Pybind11Extension(
                name='visible_residues._inner_loop',
                sources=[
                    'src/_inner_loop.cc',
                ],
                include_dirs=[
                    'src/vendored/eigen',
                ],
                cxx_std=14,
            ),
        ],
)



import macromol_voxelize as mmvox
import macromol_dataframe as mmdf
import numpy as np

atoms = mmdf.read_asymmetric_unit('1lz1.cif')
atoms = mmvox.set_atom_radius_A(atoms, 0.5)
atoms = mmvox.set_atom_channels_by_element(atoms, [['C'], ['N'], ['O'], ['S']])

img_params = mmvox.ImageParams(
        channels=4,
        grid=mmvox.Grid(
            length_voxels=15,
            resolution_A=1,
        ),
)

img, img_atoms = mmvox.image_from_atoms(atoms, img_params)
np.save('1lz1_15A.npy', img)


